///
/// @file
/// @details An entity within the ludumdare34 project.
///
/// <!-- Copyright (c) 2015 Tim Beaudet - All Rights Reserved -->
///-----------------------------------------------------------------------------------------------------------------///

#include "fire_entity.h"
#include "player_entity.h"

const tbGame::GameTimer kMaximumDelayIgnite(2500);

//--------------------------------------------------------------------------------------------------------------------//

FireEntity::FireEntity(const tbMath::Vector2& position, const FireStage& stage) :
	CollisionEntity("FireEntity", tbMath::Unstable::BoundingVolume(24.0f, 16.0f), tbMath::Vector2(0.0f, -8.0f)),
	mSprite(tbGraphics::theSpriteManager.GetAnimatedSprite("fire_sheet", "fire")),
	mDebrisSprite(tbGraphics::theSpriteManager.GetSprite("fire_sheet", "fire_debris")),
	mHasTriggered(false),
	mFireResetHackTimer(50),
	mFireStage(stage)
{
	SetFireStage(mFireStage);
	SetPosition(position);

	if (kFireSmoking == stage || kFireBlazing == stage)
	{
		OnTriggerFire();
	}

	mDebrisSprite.SetOrigin(tbGraphics::kAnchorBottomCenter);
	AddGraphic(&mDebrisSprite);

	mSprite.SetDepth(0.8f);
	mSprite.SetOrigin(tbGraphics::kAnchorBottomCenter);
	mSprite.SetPosition(0.0f, 32.0f); //move down into the ground for Daniel to work art magic
	AddGraphic(&mSprite);
}

//--------------------------------------------------------------------------------------------------------------------//

FireEntity::~FireEntity(void)
{
}

//--------------------------------------------------------------------------------------------------------------------//

void FireEntity::ChangeFireLevelTo(const FireStage& fireStage)
{
	SetFireStage(fireStage);
}

//--------------------------------------------------------------------------------------------------------------------//

void FireEntity::NextFireStage(void)
{
	if (kFireDebris == mFireStage)
	{
		SetFireStage(kFireSmoking);
		mStageIncreaseTimer = kMaximumDelayIgnite;
		tbAudio::theAudioManager.PlayEvent("audio_events", "debris_heating");
	}
	else if (kFireSmoking == mFireStage)
	{
		SetFireStage(kFireEmbers);
		mStageIncreaseTimer = kMaximumDelayIgnite;		
		tbAudio::theAudioManager.PlayEvent("audio_events", "embers_crackling");
	}
	else if (kFireEmbers == mFireStage)
	{
		SetFireStage(kFireBlazing);
		mStageIncreaseTimer = kMaximumDelayIgnite;		
		tbAudio::theAudioManager.PlayEvent("audio_events", "embers_igniting");
	}
	else if (kFireBlazing == mFireStage)
	{		
		mStageIncreaseTimer = tbGame::GameTimer::kZero;
	}
}

//--------------------------------------------------------------------------------------------------------------------//

void FireEntity::SetFireStage(const FireStage& fireStage)
{
	mFireStage = fireStage;
	if (kFireDebris == mFireStage)
	{
		mSprite.PlayAnimation("fire_debris", true, true, kAnimationFrameRate);
	}
	if (kFireSmoking == mFireStage)
	{
		mSprite.PlayAnimation("smoking", true, true, kAnimationFrameRate);
	}
	if (kFireEmbers == mFireStage)
	{
		mSprite.PlayAnimation("embers", true, true, kAnimationFrameRate);
	}
	if (kFireBlazing == mFireStage)
	{
		mSprite.PlayAnimation("blazing", true, true, kAnimationFrameRate);
	}

	if (kFireDebris == fireStage)
	{	//Bigger hitbox to ignite from fire left and right.
		SetBoundingVolume(tbMath::Unstable::BoundingVolume(48.0f, 16.0f), tbMath::Vector2(0.0f, -8.0f));
		mSprite.SetVisible(false);
	}
	else
	{
		SetBoundingVolume(tbMath::Unstable::BoundingVolume(24.0f, 16.0f), tbMath::Vector2(0.0f, -8.0f));
		mSprite.SetVisible(true);
	}
}

//--------------------------------------------------------------------------------------------------------------------//

void FireEntity::WaterOnFire(void)
{
	if (kFireBlazing == mFireStage)
	{
		tbAudio::theAudioManager.PlayEvent("audio_events", "water_on_fire");
		ChangeFireLevelTo(kFireEmbers);
		mStageIncreaseTimer = tbGame::GameTimer::kZero;
		mHasTriggered = false;
	}
}

//--------------------------------------------------------------------------------------------------------------------//

void FireEntity::OnTriggerFire(void)
{
	if (true == mHasTriggered)
	{
		return;
	}

	mStageIncreaseTimer = kMaximumDelayIgnite;
	mHasTriggered = true;

	if (kFireDebris == mFireStage)
	{
		tbAudio::theAudioManager.PlayEvent("audio_events", "debris_heating");
	}
	else if (kFireEmbers == mFireStage)
	{
		tbAudio::theAudioManager.PlayEvent("audio_events", "embers_crackling");
	}
}

//--------------------------------------------------------------------------------------------------------------------//

void FireEntity::OnAdded(void)
{
	CollisionEntity::OnAdded();
}

//--------------------------------------------------------------------------------------------------------------------//

void FireEntity::OnRemoved(void)
{
	CollisionEntity::OnRemoved();
}

//--------------------------------------------------------------------------------------------------------------------//

void FireEntity::OnSimulate(void)
{
	CollisionEntity::OnSimulate();

	mFireResetHackTimer.DecrementStep();

	if (false == mStageIncreaseTimer.IsZero())
	{
		if (true == mStageIncreaseTimer.DecrementStep())
		{
			NextFireStage();
		}
	}

	mSprite.SetOrigin(tbGraphics::kAnchorBottomCenter);
	mDebrisSprite.SetOrigin(tbGraphics::kAnchorBottomCenter);
}

//--------------------------------------------------------------------------------------------------------------------//

void FireEntity::OnUpdate(const float deltaTime)
{
	CollisionEntity::OnUpdate(deltaTime);
}

//--------------------------------------------------------------------------------------------------------------------//

void FireEntity::OnRender(void) const
{
	CollisionEntity::OnRender();
}

//--------------------------------------------------------------------------------------------------------------------//

void FireEntity::OnCollideWith(tbGame::Entity& otherEntity)
{
	CollisionEntity::OnCollideWith(otherEntity);

	if (false == mFireResetHackTimer.IsZero())
	{
		return;
	}

	if ("PlayerEntity" == otherEntity.GetEntityType())
	{
		PlayerEntity* player = dynamic_cast<PlayerEntity*>(&otherEntity);
		tb_error_if(nullptr == player , "Expected player to be an object of type PlayerEntity");

		if (kFireBlazing == mFireStage)
		{
			player->OnBurnToDeath();
		}
		if (kFireEmbers == mFireStage)
		{
			OnTriggerFire();
		}
	}

	if ("FireEntity" == otherEntity.GetEntityType())
	{
		FireEntity* fire = dynamic_cast<FireEntity*>(&otherEntity);
		tb_error_if(nullptr == fire, "Expected fire to be an object of type FireEntity");

		if (fire != this && kFireDebris == mFireStage && kFireBlazing == fire->GetFireStage())
		{
			OnTriggerFire();
		}
	}
}

//--------------------------------------------------------------------------------------------------------------------//
