///
/// @file
/// @details A way to store and set results, like current/best score for the PracticeProject project.
///
/// <!-- Copyright (c) 2015 Tim Beaudet - All Rights Reserved -->
///------------------------------------------------------------------------------------------------------------------///

#include "game_music.h"
#include "turtle_brains/tb_audio_kit.h"
#include "turtle_brains/tb_game_kit.h"

#include <list>

namespace GameMusic
{
	struct Music
	{
		tbAudio::AudioController mController;
		tbGame::GameTimer mFadeInTimer;
		tbGame::GameTimer mFadeOutTimer;
		tbCore::tbString mMusicEventName;
		float mOriginalVolume;
		bool mHasPlayed;
	};

	Music sGameMusic[kNumberOfTracks];
	size_t sCurrentTrack(kNumberOfTracks);

	const tbGame::GameTimer kFadeInTime(500);
	const tbGame::GameTimer kFadeOutTime(500);

	std::list<MusicTrack> mQueuedTracks;

}; /* namespace GameMusic */

//--------------------------------------------------------------------------------------------------------------------//

void GameMusic::InitializeMusicTracks(void)
{
	sGameMusic[kGameplayTrack].mMusicEventName = "music";
	sGameMusic[kCreditsTrack].mMusicEventName = "credits_music";

	for (size_t index = 0; index < kNumberOfTracks; ++index)
	{
	//	sGameMusic[index].mController = tbAudio::theAudioManager.PlayEvent("audio_events", sGameMusic[index].mMusicEventName);
	//	sGameMusic[index].mController.Stop();
	//	sGameMusic[index].mController.SetVolume(0.0f);

		sGameMusic[index].mHasPlayed = false;
		sGameMusic[index].mOriginalVolume = 1.0f;
		sGameMusic[index].mFadeInTimer = tbGame::GameTimer::kZero;
		sGameMusic[index].mFadeOutTimer = tbGame::GameTimer::kZero;
	}
}

//--------------------------------------------------------------------------------------------------------------------//

void GameMusic::CleanupMusicTracks(void)
{
}

//--------------------------------------------------------------------------------------------------------------------//

void GameMusic::UpdateMusicTracks(void)
{
	for (size_t index = 0; index < kNumberOfTracks; ++index)
	{
		Music& musicTrack(sGameMusic[index]);
		tb_error_if(false == musicTrack.mFadeInTimer.IsZero() && false == musicTrack.mFadeOutTimer.IsZero(), "Exptected only one timer to be active at a given time.");

		if (false == musicTrack.mFadeInTimer.IsZero())
		{
			if (true == musicTrack.mFadeInTimer.DecrementStep())
			{
				musicTrack.mController.SetVolume(1.0f * musicTrack.mOriginalVolume);
			}
			else
			{
				musicTrack.mController.SetVolume((1.0f - musicTrack.mFadeInTimer.GetPercentageOf(kFadeInTime)) * musicTrack.mOriginalVolume);
			}
		}

		if (false == musicTrack.mFadeOutTimer.IsZero())
		{
			if (true == musicTrack.mFadeOutTimer.DecrementStep())
			{
				musicTrack.mController.SetVolume(0.0f);
				musicTrack.mController.Stop();
			}
			else
			{
				musicTrack.mController.SetVolume(musicTrack.mFadeOutTimer.GetPercentageOf(kFadeOutTime) * musicTrack.mOriginalVolume);
			}
		}
	}

	if (kNumberOfTracks != sCurrentTrack && true == sGameMusic[sCurrentTrack].mController.IsComplete())
	{
		if (false == mQueuedTracks.empty())
		{
			SetMusicTrack(mQueuedTracks.front());
			mQueuedTracks.push_back(mQueuedTracks.front());
			mQueuedTracks.pop_front();
		}
		else
		{
			SetMusicTrack(kNumberOfTracks);
		}
	}
}

//--------------------------------------------------------------------------------------------------------------------//

void GameMusic::SetMusicTrack(const MusicTrack& musicTrack, bool clearQueuedTracks)
{
	if (true == clearQueuedTracks)
	{
		mQueuedTracks.clear();
	}

	if (kNumberOfTracks == musicTrack)
	{
		for (size_t index = 0; index < kNumberOfTracks; ++index)
		{
			Music& musicTrack(sGameMusic[index]);
			musicTrack.mFadeInTimer = tbGame::GameTimer::kZero;
			musicTrack.mFadeOutTimer = kFadeOutTime;
		}

		sCurrentTrack = kNumberOfTracks;
		return;
	}
	
	if (sCurrentTrack == musicTrack)
	{	//Special case attempting to set the current track, just continue playing.
		return;
	}

	if (kNumberOfTracks != sCurrentTrack)
	{	//Start fading out the current / old track before setting up the new track to fade in.
		Music& currentMusicTrack(sGameMusic[sCurrentTrack]);
		currentMusicTrack.mFadeInTimer = tbGame::GameTimer::kZero;
		currentMusicTrack.mFadeOutTimer = kFadeOutTime;
	}

	Music& newMusicTrack(sGameMusic[musicTrack]);
	newMusicTrack.mController.Stop(); //In the event the old audio is still playing, kill it.  Unlikely if everything else went to plan.
	newMusicTrack.mController = tbAudio::theAudioManager.PlayEvent("audio_events", newMusicTrack.mMusicEventName);
	if (false == newMusicTrack.mHasPlayed)
	{
		newMusicTrack.mOriginalVolume = newMusicTrack.mController.GetVolume();
		newMusicTrack.mHasPlayed = true;
	}
	newMusicTrack.mController.SetVolume(0.0f);
	newMusicTrack.mFadeInTimer = kFadeInTime;
	newMusicTrack.mFadeOutTimer = tbGame::GameTimer::kZero;
	sCurrentTrack = musicTrack;
}

//--------------------------------------------------------------------------------------------------------------------//

void GameMusic::QueueMusicTrack(const MusicTrack& musicTrack)
{
	mQueuedTracks.push_back(musicTrack);
}

//--------------------------------------------------------------------------------------------------------------------//
