///
/// @file
/// @details This scene is the gameplay scene for the ludumdare34 project.
///
/// <!-- Copyright (c) 2015 Tim Beaudet - All Rights Reserved -->
///-----------------------------------------------------------------------------------------------------------------///

#include "gameplay_scene.h"
#include "scene_manager.h"
#include "score_display.h"
#include "game_results.h"

#include "gif_export/gif_exporter.h"

#include "goal_entity.h"
#include "fire_entity.h"
#include "tree_entity.h"
#include "pickup_entity.h"
#include "player_entity.h"
#include "slanted_entity.h"
#include "level_name_entity.h"

#ifdef developer_build
//const tbCore::tbString kFirstLevelFile("data/levels/final_level.json");
//const tbCore::tbString kFirstLevelFile("data/levels/debugLevel.json");
//const tbCore::tbString kFirstLevelFile("data/levels/first_level.json");
const tbCore::tbString kFirstLevelFile("data/levels/level08.json");
#else
const tbCore::tbString kFirstLevelFile("data/levels/first_level.json");
#endif

//--------------------------------------------------------------------------------------------------------------------//

GameplayScene::GameplayScene(void) :
	FadingScene(),
	mBackdropSprite(tbGraphics::theSpriteManager.GetSpriteFromFile("data/backdrop_level_01.png")),
	mCurrentLevel(kFirstLevelFile),
	mLoadLevel(kFirstLevelFile)
{
}

//--------------------------------------------------------------------------------------------------------------------//

GameplayScene::~GameplayScene(void)
{
}

//--------------------------------------------------------------------------------------------------------------------//

void GameplayScene::OnSimulate(void)
{
	FadingScene::OnSimulate();
}

//--------------------------------------------------------------------------------------------------------------------//

void GameplayScene::OnUpdate(const float deltaTime)
{
	if (kBlackScreen == GetFadeState())
	{
		OnLoadLevel(mLoadLevel);
		FadeToGame();
	}

	FadingScene::OnUpdate(deltaTime);

#ifdef developer_build
	if (true == tbApplication::Input::IsKeyReleased(tbApplication::tbKeyE))
	{
		const tbCore::tbString levelFile(mTileSystem.GetMapPropertyAsString("next_level"));
		SignalLevelLoad(levelFile);
	}
#endif /* developer_build */

	if (true == tbApplication::Input::IsKeyReleased(tbApplication::tbKeyEscape))
	{	
		SignalLevelLoad("title");
	}
}

//--------------------------------------------------------------------------------------------------------------------//

void GameplayScene::OnRender(void) const
{
	FadingScene::OnRender();
}

//--------------------------------------------------------------------------------------------------------------------//

void GameplayScene::OnOpen(void)
{
	FadingScene::OnOpen();

	OnLoadLevel(mCurrentLevel);
}

//--------------------------------------------------------------------------------------------------------------------//

void GameplayScene::OnClose(void)
{
	FadingScene::OnClose();
}

//--------------------------------------------------------------------------------------------------------------------//

void GameplayScene::SignalLevelLoad(const tbCore::tbString& levelFile)
{
	mLoadLevel = levelFile;
	FadeToBlack();
}

//--------------------------------------------------------------------------------------------------------------------//

void GameplayScene::SignalLevelReset(void)
{
	SignalLevelLoad(mCurrentLevel);
}

//--------------------------------------------------------------------------------------------------------------------//

void GameplayScene::OnLoadLevel(const tbCore::tbString& levelFile)
{
	if ("credits" == levelFile)
	{
		mCurrentLevel = kFirstLevelFile;
		theSceneManager->ChangeToScene(kCreditsScene);
		return;
	}
	if ("title" == levelFile)
	{
		theSceneManager->ChangeToScene(kTitleScene);
		return;
	}

	mCurrentLevel = levelFile;

	tbGame::Extensions::LoadTileSystemFromTiled(mTileSystem, levelFile);	
	mTileSystem.UpdateColliderInformation();

	{	//Turn collision off just in case there is an extra frame with the entities in existance.
		std::list<tbGame::EntityInterface*> entities = GetAllEntities();
		for (tbGame::EntityInterface* entity : entities)
		{
			entity->SetCollidable(false);
		}
	}

	RemoveEntities(); //Remove ALL Entities.

	//Add backdrop first as it is behind all entities.
	AddEntity(new BackdropEntity(mTileSystem));	


#ifdef developer_build
	static tbGraphics::Text* levelText = new tbGraphics::Text(mCurrentLevel);
	levelText->SetText(mCurrentLevel);
	levelText->SetPosition(1260.0f - levelText->GetWidth(), 20.0f);
	levelText->SetDepth(1.0f);
	levelText->SetColor(tbGraphics::Color(0xFFFF0000));

	RemoveGraphic(levelText);
	AddGraphic(levelText);
#endif /* developer_build */

	const tbGame::AboutTileContainer& goalTiles = mTileSystem.FindTilesWithPropertyEquals("player_end", true);
	tb_error_if(goalTiles.empty(), "Expected level to have a tile with property player_end.");
	for (const tbGame::AboutTile& info : goalTiles)
	{
		tbMath::Vector2 position(info.mCenterX, info.mTop - info.mHeight);
		tbCore::tbString nextLevelFile = mTileSystem.GetMapPropertyAsString("next_level");
		AddEntity(new GoalEntity(position, nextLevelFile));		
	}

	const tbGame::AboutTileContainer& plantableTiles = mTileSystem.FindTilesWithProperty("tree");
	for (const tbGame::AboutTile& info : plantableTiles)
	{
		tbMath::Vector2 position(info.mCenterX, info.mTop);

		int tree = info.mPropertyValue.AsInteger();
		TreeEntity::GrowthStage treeSize(TreeEntity::kSeedling);
		if (1 == tree) { treeSize = TreeEntity::kLittleTree; }
		if (2 == tree) { treeSize = TreeEntity::kNormalTree; }
		if (3 == tree) { treeSize = TreeEntity::kLargeTree;  }
		
		AddEntity(new TreeEntity(position, treeSize));
	}

	const tbGame::AboutTileContainer& slantedTiles = mTileSystem.FindTilesWithProperty("slanted");
	for (const tbGame::AboutTile& info : slantedTiles)
	{
		tbMath::Vector2 position(info.mCenterX, info.mTop + info.mHeight);

		int slanted = info.mPropertyValue.AsInteger();
		bool isSlantedUpRight(false);
		if (0 == slanted) { isSlantedUpRight = false;  }
		if (1 == slanted) { isSlantedUpRight = true; }

		AddEntity(new SlantedEntity(position, isSlantedUpRight));
	}

	const tbGame::AboutTileContainer& waterPickups = mTileSystem.FindTilesWithPropertyEquals("water_can", true);
	for (const tbGame::AboutTile& info : waterPickups)
	{
		tbMath::Vector2 position(info.mCenterX, info.mCenterY - info.mHeight);
		AddEntity(new PickupEntity(position));
	}

	//Add player almost last because it is in front of all entities.
	AddEntity(new PlayerEntity(mTileSystem));

	//Add fire last because it is in front of player.
	const tbGame::AboutTileContainer& fireTiles = mTileSystem.FindTilesWithProperty("fire");
	for (const tbGame::AboutTile& info : fireTiles)
	{
		tbMath::Vector2 position(info.mCenterX, info.mTop);

		int fire = info.mPropertyValue.AsInteger();
		FireEntity::FireStage fireSize(FireEntity::kFireSmoking);
		if (3 == fire) { fireSize = FireEntity::kFireDebris; }
		if (0 == fire) { fireSize = FireEntity::kFireEmbers; }
		if (1 == fire) { fireSize = FireEntity::kFireBlazing; }

		AddEntity(new FireEntity(position, fireSize));
	}

	const tbGame::AboutTileContainer& goldenPickups = mTileSystem.FindTilesWithProperty("golden_water_can");
	for (const tbGame::AboutTile& info : goldenPickups)
	{
		tbMath::Vector2 position(info.mCenterX, info.mTop);
		AddEntity(new PickupEntity(position, true));
	}

	AddEntity(new LevelNameEntity(mTileSystem.GetMapPropertyAsString("name")));
}

//--------------------------------------------------------------------------------------------------------------------//
//--------------------------------------------------------------------------------------------------------------------//
//--------------------------------------------------------------------------------------------------------------------//

BackdropEntity::BackdropEntity(tbGame::TileSystem& tileSystem) :
	tbGame::Entity("BackdropEntity"),
	mTileSystem(tileSystem),
	mBackdropSprite(tbGraphics::theSpriteManager.GetSpriteFromFile("data/backdrop_level_01.png"))
{

	tbCore::tbString backdropFile(mTileSystem.GetMapPropertyAsString("backdrop"));
	//if (true == backdropFile.empty())
	//{
	//	backdropFile = tb_string("data/backdrop_level_01.png");
	//}
	mBackdropSprite = tbGraphics::theSpriteManager.GetSpriteFromFile(backdropFile);
	mBackdropSprite.SetDepth(-1.0f);
	AddGraphic(&mBackdropSprite);

	mTileSystem.SetDepth(-0.5f);
	AddGraphic(&mTileSystem);
}

//--------------------------------------------------------------------------------------------------------------------//

BackdropEntity::~BackdropEntity(void)
{
}

//--------------------------------------------------------------------------------------------------------------------//
