///
/// @file
/// @details An entity within the ludumdare34 project.
///
/// <!-- Copyright (c) 2015 Tim Beaudet - All Rights Reserved -->
///-----------------------------------------------------------------------------------------------------------------///

#include "player_entity.h"
#include "goal_entity.h"
#include "tree_entity.h"
#include "fire_entity.h"
#include "pickup_entity.h"
#include "slanted_entity.h"
#include "scene_manager.h"
#include "gameplay_scene.h"
#include "game_music.h"

//--------------------------------------------------------------------------------------------------------------------//

const tbGame::GameTimer kSeedsUseCooldownTime(1000);
const tbGame::GameTimer kWaterUseCooldownTime(1000);
const tbGame::GameTimer kClimbMoveTime(500);
const tbGame::GameTimer kClimbDelayTime(0);
const tbGame::GameTimer kDeathDelayResetTime(2500);

const float kFallDistanceToDeath(32.0f * 2.5f); 

//--------------------------------------------------------------------------------------------------------------------//
//--------------------------------------------------------------------------------------------------------------------//
//--------------------------------------------------------------------------------------------------------------------//

class WinningFadeBehavior : public tbGame::EntityBehaviorInterface
{
public:
	WinningFadeBehavior(tbGame::EntityInterface& entity, tbGraphics::AnimatedSprite& playerSprite, tbGraphics::Sprite& fadeSprite) :
		EntityBehaviorInterface(entity),
		mPlayerSprite(playerSprite),
		mFadeSprite(fadeSprite),
		mFadeTimer(kMaximumFadeTime),
		mFirstFrame(true),
		mHasDelayed(false)
	{
	}

	virtual void OnSimulate(void)
	{
		if (true == mFirstFrame)
		{
			tbAudio::theAudioManager.PlayEvent("audio_events", "collect_golden_water_can");
			mFirstFrame = false;
		}

		mFadeSprite.SetVisible(true);

		mPlayerSprite.PlayAnimation("pickup_gold", false, true, kAnimationFrameRate);

		if (true == mFadeTimer.DecrementStep())
		{
			mFadeSprite.SetColor(tbGraphics::Color(0xBB000000));

			if (false == mHasDelayed)
			{
				mPlayerSprite.StopAnimation();
				mEntity.PushBehavior(new tbGame::DelayBehavior(mEntity, 0.5f));
				mHasDelayed = true;
			}
			else
			{
				mEntity.PopBehavior();
			}
		}
		else
		{
			float tween = 1.0f - mFadeTimer.GetPercentageOf(kMaximumFadeTime);
			tbGraphics::Color color;
			color.SetColor(tween, tbGraphics::Color(0x00000000), tbGraphics::Color(0xBB000000));
			mFadeSprite.SetColor(color);

			//Wait...
			mPlayerSprite.StopAnimation();
		}
	}
private:
	static const tbGame::GameTimer kMaximumFadeTime;
	tbGame::GameTimer mFadeTimer;
	tbGraphics::AnimatedSprite& mPlayerSprite;
	tbGraphics::Sprite& mFadeSprite;
	bool mFirstFrame;
	bool mHasDelayed;
};

const tbGame::GameTimer WinningFadeBehavior::kMaximumFadeTime(1500);

class WinningAnimationBehavior : public tbGame::EntityBehaviorInterface
{
public:
	WinningAnimationBehavior(tbGame::EntityInterface& entity, tbGraphics::AnimatedSprite& animatedSprite, tbGame::Entity* removeEntity) :
		EntityBehaviorInterface(entity),
		mAnimatedSprite(animatedSprite),
		mRemoveEntity(removeEntity)
	{
	}

	virtual void OnResume(void)
	{	//TODO: TIM: TurtleBrains doesn't call OnResume it seems...
		mAnimatedSprite.PlayAnimation("reading", false, true, kAnimationFrameRate);
	}

	virtual void OnSimulate(void)
	{
		if (false == mAnimatedSprite.IsPlayingAnimation("pickup_gold"))
		{
			if (nullptr != mRemoveEntity)
			{
				mRemoveEntity->SetVisible(false);
				mRemoveEntity->GetEntityManager()->RemoveEntity(mRemoveEntity);
			}
			mAnimatedSprite.PlayAnimation("pickup_gold", false, true, kAnimationFrameRate);
		}

		if (false == mAnimatedSprite.IsAnimationPlaying())
		{
			mAnimatedSprite.PlayAnimation("pickup_gold_loop", true, true, kAnimationFrameRate);
			mEntity.PopBehavior();
		}
	}

private:
	tbGraphics::AnimatedSprite& mAnimatedSprite;
	tbGame::Entity* mRemoveEntity;
};

class EndLevelBehavior : public tbGame::EntityBehaviorInterface
{
public:
	EndLevelBehavior(tbGame::EntityInterface& entity) :
		EntityBehaviorInterface(entity)
	{
	}

	virtual void OnSimulate(void)
	{
		GameplayScene* gameplayScene = dynamic_cast<GameplayScene*>(&theSceneManager->GetScene(kGameplayScene));
		tb_error_if(nullptr == gameplayScene, "Expected gameplayScene to be a GameplayScene object!");
		gameplayScene->SignalLevelLoad("credits");

		mEntity.PopBehavior();
	}
};

//--------------------------------------------------------------------------------------------------------------------//
//--------------------------------------------------------------------------------------------------------------------//
//--------------------------------------------------------------------------------------------------------------------//

PlayerEntity::PlayerEntity(const tbGame::TileSystem& tileSystem) :
	CollisionEntity("PlayerEntity", tbMath::Unstable::BoundingVolume(28.0f, 60.0f), tbMath::Vector2(0.0f, -30.0f)),
	mTileSystem(tileSystem),
	mSprite(tbGraphics::theSpriteManager.GetAnimatedSprite("player_sheet", "player")),
	mWinningScreenFade(tbGraphics::theSpriteManager.GetSpriteFromFile("data/blank_square.png")),
	mIsAlive(true),
	mIsOnSlant(false),
	mHasWonGame(false),
	mHasWonLevel(false),
	mIsFacingLeft(false),
	mIsMovingLeft(false),
	mIsMovingRight(false),
	mIsClimbingUp(false),
	mIsClimbingDown(false),
	mIsUsingWater(false),
	mIsUsingSeeds(false),
	mIsOnTree(false),
	mSeedsUseTimer(0),
	mWaterUseTimer(0),
	mRemainingSeeds(3),
	mRemainingWater(3),
	mClimbMove(kClimbNone),
	mClimbMoveTimer(0),
	mDeathDelayResetTimer(kDeathDelayResetTime)
{
	ResetToStart();

	mWinningScreenFade.SetOrigin(tbGraphics::kAnchorCenter);
	mWinningScreenFade.SetScale(100.0f, 100.0f);
	//mWinningScreenFade.SetColor(tbGraphics::Color::kBlack);
	mWinningScreenFade.SetColor(tbGraphics::Color(0xA0000000));
	mWinningScreenFade.SetDepth(0.6f);
	mWinningScreenFade.SetVisible(false);
	AddGraphic(&mWinningScreenFade);

	mSprite.SetPosition(0.0f, 2.0f);
	mSprite.SetOrigin(tbGraphics::kAnchorBottomCenter);
	mSprite.SetDepth(0.7f);
	AddGraphic(&mSprite);
}

//--------------------------------------------------------------------------------------------------------------------//

PlayerEntity::~PlayerEntity(void)
{
}

//--------------------------------------------------------------------------------------------------------------------//

void PlayerEntity::OnAdded(void)
{
	CollisionEntity::OnAdded();
}

//--------------------------------------------------------------------------------------------------------------------//

void PlayerEntity::OnRemoved(void)
{
	CollisionEntity::OnRemoved();
}

//--------------------------------------------------------------------------------------------------------------------//

void PlayerEntity::OnSimulate(void)
{
	CollisionEntity::OnSimulate();

	if (false == mIsAlive)
	{
		if (false == mDeathDelayResetTimer.IsZero())
		{
			if (true == mDeathDelayResetTimer.DecrementStep())
			{
				GameplayScene* gameplayScene = dynamic_cast<GameplayScene*>(&theSceneManager->GetScene(kGameplayScene));
				tb_error_if(nullptr == gameplayScene, "Expected gameplayScene to be a GameplayScene object!");
				gameplayScene->SignalLevelReset();
			}
		}
	}

	if (true == mIsUsingSeeds)
	{
		//mIsUsingSeeds = false;
		//--mRemainingSeeds;
		//
		//mSeedsUseTimer = kSeedsUseCooldownTime;
		
		OnSuicidalDeath();
	}
	mSeedsUseTimer.DecrementStep();

	if (false == mHasWonLevel && true == mIsAlive && true == mIsUsingWater)
	{
		mIsUsingWater = false;
		--mRemainingWater;
		tbAudio::theAudioManager.PlayEvent("audio_events", "water_tree");
		if (true == mIsFacingLeft)
		{
			if (false == mSprite.IsPlayingAnimation("water_left")) { mSprite.PlayAnimation("water_left", true, true, kAnimationFrameRate); }
		}
		else
		{
			if (false == mSprite.IsPlayingAnimation("water_right")) { mSprite.PlayAnimation("water_right", true, true, kAnimationFrameRate); }
		}

		mWaterUseTimer = kWaterUseCooldownTime;
		SimulateUsingWater();
	}
	mWaterUseTimer.DecrementStep();
	
	SimulateMovement();

	if (false == mHasWonLevel && true == mIsAlive)
	{
		tbMath::Vector2 currentPosition(GetPosition());
		const float fallDistance = (currentPosition.y - mLastSafeHeight.y);

		if (false == mIsOnTree && false == IsOnGround())
		{
			if (fallDistance > 5.0f)
			{
				if (true == mIsFacingLeft)
				{
					if (false == mSprite.IsPlayingAnimation("fall_left")) { mSprite.PlayAnimation("fall_left", true, true, kAnimationFrameRate); }
				}
				else
				{
					if (false == mSprite.IsPlayingAnimation("fall_right")) { mSprite.PlayAnimation("fall_right", true, true, kAnimationFrameRate); }
				}
			}
		}
		else
		{			
			if (fallDistance > kFallDistanceToDeath)
			{
				OnFallDeath();
			}
			mLastSafeHeight = currentPosition;
		}
	}

	if (false == mHasWonLevel && true == mIsAlive && true == IsOnGround() && kClimbNone == mClimbMove && mWaterUseTimer.IsZero())
	{
		if (true == IsMoving())
		{
			if (true == mIsFacingLeft)
			{
				if (false == mSprite.IsPlayingAnimation("run_left")) { mSprite.PlayAnimation("run_left", true, true, kAnimationFrameRate); }
			}
			else
			{
				if (false == mSprite.IsPlayingAnimation("run_right")) { mSprite.PlayAnimation("run_right", true, true, kAnimationFrameRate); }
			}
		}
		else
		{
			if (true == mIsFacingLeft)
			{
				if (false == mSprite.IsPlayingAnimation("stand_left")) { mSprite.PlayAnimation("stand_left", true, true, kAnimationFrameRate); }
			}
			else
			{
				if (false == mSprite.IsPlayingAnimation("stand_right")) { mSprite.PlayAnimation("stand_right", true, true, kAnimationFrameRate); }
			}
		}
	}

	mIsOnSlant = false;
	mIsOnTree = false;

	mSprite.SetOrigin(tbGraphics::kAnchorBottomCenter);

	tbMath::Vector2 position(GetPosition());
	bool outOfBounds((position.x < -20.0f || position.x > 1300.0f || position.y < -70.0f || position.y > 790.0f) ? true : false);
	if (false == mHasWonLevel && true == mIsAlive && false == mHasWonLevel && true == outOfBounds)
	{
		mSprite.SetVisible(false);
		OnSuicidalDeath();
	}
}

//--------------------------------------------------------------------------------------------------------------------//

void PlayerEntity::OnUpdate(const float deltaTime)
{
	CollisionEntity::OnUpdate(deltaTime);

	mIsMovingLeft = (true == tbApplication::Input::IsKeyDown(tbApplication::tbKeyLeft) || true == tbApplication::Input::IsKeyDown(tbApplication::tbKeyA));
	mIsMovingRight = (true == tbApplication::Input::IsKeyDown(tbApplication::tbKeyRight) || true == tbApplication::Input::IsKeyDown(tbApplication::tbKeyD));
	mIsClimbingUp = (true == tbApplication::Input::IsKeyDown(tbApplication::tbKeyUp) || true == tbApplication::Input::IsKeyDown(tbApplication::tbKeyW));
	mIsClimbingDown = (true == tbApplication::Input::IsKeyDown(tbApplication::tbKeyDown) || true == tbApplication::Input::IsKeyDown(tbApplication::tbKeyS));

	bool isOnGround = IsOnGround();
	bool isMoving = IsMoving();

	//if (false == mIsUsingSeeds && mSeedsUseTimer.IsZero() && mRemainingSeeds > 0 && true == isOnGround && false == isMoving && kClimbNone == mClimbMove)
	//{
	//	mIsUsingSeeds = (true == tbApplication::Input::IsKeyPressed(tbApplication::tbKeyZ) || true == tbApplication::Input::IsKeyPressed(tbApplication::tbKeyJ));
	//}

	if (false == mHasWonLevel && true == mIsAlive && true == isOnGround && kClimbNone == mClimbMove)
	{
		if (true == tbApplication::Input::IsKeyPressed(tbApplication::tbKeyZ) || true == tbApplication::Input::IsKeyPressed(tbApplication::tbKeyJ))
		{
			OnSuicidalDeath();
		}
	}

	//if (true == mIsAlive && false == mIsUsingWater && mWaterUseTimer.IsZero() && mRemainingWater > 0 && true == isOnGround && false == isMoving && kClimbNone == mClimbMove)
	if (false == mHasWonLevel && true == mIsAlive && false == mIsUsingWater && mWaterUseTimer.IsZero() && mRemainingWater > 0 && true == isOnGround && kClimbNone == mClimbMove)
	{
		mIsUsingWater = (true == tbApplication::Input::IsKeyPressed(tbApplication::tbKeyX) || true == tbApplication::Input::IsKeyPressed(tbApplication::tbKeyK));
	}

	//mFootStepSoundDelay.DecrementStep();
	//if (true == mFootStepSoundDelay.IsZero() && true == mSprite.IsPlayingAnimation("run"))
	if (true == mSprite.IsPlayingAnimation("run_left") || true == mSprite.IsPlayingAnimation("run_right"))
	{	//3 and 7
		if (3 == mSprite.GetFrameIndex() || 7 == mSprite.GetFrameIndex())
		{
			tbAudio::theAudioManager.PlayEvent("audio_events", "run_footstep");
		}
	}

	mSprite.SetOrigin(tbGraphics::kAnchorBottomCenter);
}

//--------------------------------------------------------------------------------------------------------------------//

void PlayerEntity::OnRender(void) const
{
	CollisionEntity::OnRender();

	tbGraphics::Sprite waterSprite(tbGraphics::theSpriteManager.GetSpriteFromFile("data/water_icon.png"));
	waterSprite.SetOrigin(tbGraphics::kAnchorTopLeft);

	tbGraphics::Sprite emptyWaterSprite(tbGraphics::theSpriteManager.GetSpriteFromFile("data/water_icon_empty.png"));
	emptyWaterSprite.SetOrigin(tbGraphics::kAnchorTopLeft);

	int water = 0;
	for ( ; water < mRemainingWater; ++water)
	{
		waterSprite.SetPosition(20.0f + (36.0f * water), 20.0f);
		waterSprite.Render();
	}

	for (; water < 6; ++water)
	{
		emptyWaterSprite.SetPosition(20.0f + (36.0f * water), 20.0f);
		emptyWaterSprite.Render();
	}

	CollisionEntity::RenderBoundingVolume(GetWaterBoundingVolume(), GetPosition() + GetWaterBoundingOffset());
}

//--------------------------------------------------------------------------------------------------------------------//

void PlayerEntity::OnCollectWater(const int count)
{
	mRemainingWater += count;
	if (mRemainingWater > 6)
	{
		mRemainingWater = 6;
	}
}

//--------------------------------------------------------------------------------------------------------------------//

void PlayerEntity::OnCollideWith(tbGame::Entity& otherEntity)
{
	CollisionEntity::OnCollideWith(otherEntity);

	if ("TreeEntity" == otherEntity.GetEntityType())
	{
		mIsOnTree = true;
	}

	if ("PickupEntity" == otherEntity.GetEntityType())
	{
		PickupEntity* pickupEntity = dynamic_cast<PickupEntity*>(&otherEntity);
		tb_error_if(nullptr == pickupEntity, "Expected pickupEntity to be a PickupEntity object!");
		pickupEntity->SetCollidable(false);

		if (pickupEntity->IsGoldenCan())
		{
			mHasWonLevel = true;
			mHasWonGame = true;

			PushBehavior(new EndLevelBehavior(*this));
			PushBehavior(new tbGame::DelayBehavior(*this, 3.0f));
			PushBehavior(new WinningAnimationBehavior(*this, mSprite, pickupEntity));
			PushBehavior(new WinningFadeBehavior(*this, mSprite, mWinningScreenFade));
			PushBehavior(new tbGame::DelayBehavior(*this, 0.35f));
			PushBehavior(new tbGame::MoveToBehavior(*this, pickupEntity->GetPosition(), 50.0f));

			GameMusic::SetMusicTrack(GameMusic::kNumberOfTracks);
		}
		else
		{
			pickupEntity->OnCollected();
			tbAudio::theAudioManager.PlayEvent("audio_events", "collect_water");
		}
	}

	if ("GoalEntity" == otherEntity.GetEntityType())
	{
		mHasWonLevel = true;

		GoalEntity* goalEntity = dynamic_cast<GoalEntity*>(&otherEntity);
		tb_error_if(nullptr == goalEntity, "Expected goalEntity to be a GoalEntity object!");

		GameplayScene* gameplayScene = dynamic_cast<GameplayScene*>(&theSceneManager->GetScene(kGameplayScene));
		tb_error_if(nullptr == gameplayScene, "Expected gameplayScene to be a GameplayScene object!");
		gameplayScene->SignalLevelLoad(goalEntity->GetNextLevelFile());
		goalEntity->SetCollidable(false);
	}

	if ("SlantedEntity" == otherEntity.GetEntityType())
	{
		//PlayerEntity* player = dynamic_cast<PlayerEntity*>(&otherEntity);
		//tb_error_if(nullptr == player, "Expected player to be an object of type PlayerEntity");

		//const float playerX(player->GetPosition().x);
		//float heightAt = FindHeightAt(playerX);
		//player->SetPosition(playerX, heightAt);

		SlantedEntity* slant = dynamic_cast<SlantedEntity*>(&otherEntity);
		tb_error_if(nullptr == slant, "Expected slant to be an object of type SlantedEntity");

		const float playerX(GetPosition().x);
		float heightAt = slant->FindHeightAt(playerX);
		SetPosition(playerX, heightAt);
		mPreviousPosition.y = heightAt; //no Y velocity.

		mIsOnSlant = true;
	}
}

//--------------------------------------------------------------------------------------------------------------------//

bool PlayerEntity::IsMoving(void) const
{
	const tbMath::Vector2 currentPosition(GetPosition());
	tbMath::Vector2 velocity = (currentPosition - mPreviousPosition) / kFixedTime;
	return (fabs(velocity.x) < 5.0f && fabs(velocity.y) < 5.0f) ? false : true;
}

//--------------------------------------------------------------------------------------------------------------------//

bool PlayerEntity::IsOnGround(void) const
{
	if (true == mIsOnSlant) { return true; }
	if (true == mTileSystem.IsPointInSolid(GetPosition() + tbMath::Vector2(-9.0f, 1.0f))) { return true; }
	if (true == mTileSystem.IsPointInSolid(GetPosition() + tbMath::Vector2(0.0f, 1.0f))) { return true; }
	if (true == mTileSystem.IsPointInSolid(GetPosition() + tbMath::Vector2(9.0f, 1.0f))) { return true; }
	return false;
}

//--------------------------------------------------------------------------------------------------------------------//

bool PlayerEntity::IsClimbing(void) const
{
	return (true == mIsOnTree && false == IsOnGround());
}

//--------------------------------------------------------------------------------------------------------------------//

tbMath::Vector2 PlayerEntity::SimulateMovement(void)
{
	if (true == mHasWonGame || false == mIsAlive)
	{
		return GetPosition();
	}

	static tbGame::GameTimer climbShimmyTimer;
	climbShimmyTimer.IncrementStep();
	float climbShimmySine = sinf(tbGame::GameTimer::ToSeconds(climbShimmyTimer.GetElapsedTime()));

	const tbMath::Vector2 kGravityAcceleration(0.0f, 900.0f);   //Down
	const tbMath::Vector2 kRunningAcceleration(1000.0f, 0.0f);	//Right

	const tbMath::Vector2 currentPosition(GetPosition());
	tbMath::Vector2 velocity = (currentPosition - mPreviousPosition) / kFixedTime;

	tbMath::Vector2 desiredClimbingPosition(currentPosition);

	const bool isOnGround(IsOnGround());

	if (true == isOnGround && mWaterUseTimer.IsZero())
	{
		//Running Movement
		if (true == mIsMovingLeft)
		{
			mIsFacingLeft = true;
			velocity -= (kRunningAcceleration * kFixedTime);
		}
		if (true == mIsMovingRight)
		{
			mIsFacingLeft = false;
			velocity += (kRunningAcceleration * kFixedTime);
		}
	}

	if (true == isOnGround)
	{
		//if (false == mIsMovingLeft && false == mIsMovingRight)
		{
			velocity.x -= ((velocity.x * 8.0f) * kFixedTime);
		}
	}


	if (true == mIsOnTree)
	{		
		if (false == isOnGround || kClimbNone != mClimbMove)
		{
			velocity = tbMath::Vector2::kZero;
		}

		static int climbAnim = 0;
		if (kClimbNone == mClimbMove)
		{
			mStartOfClimbMove = currentPosition;

			if (true == mIsClimbingUp)
			{
				tbCore::tbString animNum = tbCore::ToString(++climbAnim % 2);
				mClimbMove = kClimbUp;
				//mSprite.PlayAnimation("climb_up", false, true, kAnimationFrameRate / 2.0f);
				//mSprite.PlayAnimation("climb_up", false, true, kAnimationFrameRate);
				mSprite.PlayAnimation("climb_up_" + animNum, false, true, kAnimationFrameRate);
				tbAudio::theAudioManager.PlayEvent("audio_events", "climb_up");
			}
			if (true == mIsClimbingDown && false == isOnGround)
			{
				tbCore::tbString animNum = tbCore::ToString(++climbAnim % 2);
				mClimbMove = kClimbDown;
				//mSprite.PlayAnimation("climb_down", false, true, kAnimationFrameRate / 2.0f);
				//mSprite.PlayAnimation("climb_down", false, true, kAnimationFrameRate);
				mSprite.PlayAnimation("climb_down_" + animNum, false, true, kAnimationFrameRate);

				tbAudio::theAudioManager.PlayEvent("audio_events", "climb_down");
			}
			if (true == mIsMovingLeft && false == isOnGround)
			{
				mClimbMove = kShimmyLeft;
				mSprite.PlayAnimation("shim_left", false, true, kAnimationFrameRate);
				tbAudio::theAudioManager.PlayEvent("audio_events", "shimmy");
			}
			if (true == mIsMovingRight && false == isOnGround)
			{
				mClimbMove = kShimmyRight;
				mSprite.PlayAnimation("shim_right", false, true, kAnimationFrameRate);
				tbAudio::theAudioManager.PlayEvent("audio_events", "shimmy");
			}

			if (kClimbNone != mClimbMove) { mClimbMoveTimer = kClimbMoveTime; }
		}
		else if (kClimbDelay == mClimbMove)
		{
			if (true == mClimbMoveTimer.DecrementStep())
			{
				mClimbMove = kClimbNone;
			}
		}
		else
		{
			tbMath::Vector2 climbMovement(0.0f, 0.0f);
			if (kClimbUp == mClimbMove) { climbMovement = tbMath::Vector2(0.0f, -16.0f); }
			if (kClimbDown == mClimbMove) { climbMovement = tbMath::Vector2(0.0f, 16.0f); }
			if (kShimmyLeft == mClimbMove) { climbMovement = tbMath::Vector2(-16.0f, 0.0f); }
			if (kShimmyRight == mClimbMove) { climbMovement = tbMath::Vector2(16.0f, 0.0f); }

			if (kClimbDown == mClimbMove || kClimbUp == mClimbMove) { mClimbMoveTimer.DecrementStep(); } //Doublespeed up/down

			if (true == mClimbMoveTimer.DecrementStep())
			{
				desiredClimbingPosition = mStartOfClimbMove + climbMovement;
				mClimbMoveTimer = kClimbDelayTime;
				mClimbMove = kClimbDelay;
			}
			else
			{
				float tween = (1.0f - mClimbMoveTimer.GetPercentageOf(kClimbMoveTime));
				if (kShimmyLeft == mClimbMove || kShimmyRight == mClimbMove)
				{	//Only do the smoothstep during shimmy sideways.
					tween = tbMath::Interpolation::SmoothStep(tbMath::Interpolation::SmoothStep(tween));
				}
				desiredClimbingPosition = mStartOfClimbMove + (climbMovement * tween);
			}
		}
	}
	else
	{
		mClimbMove = kClimbNone;
		if (false == isOnGround)
		{
			velocity += kGravityAcceleration * kFixedTime;
		}
	}
	
	tbMath::Vector2 desiredPosition(desiredClimbingPosition + (velocity * kFixedTime) + GetBoundingOffset());
	if (false == mIsOnSlant)
	{
		mTileSystem.MoveEntity(currentPosition + GetBoundingOffset(), desiredPosition, GetBoundingVolume());
	}

	mPreviousPosition = currentPosition;
	const tbMath::Vector2 finalPosition(desiredPosition - GetBoundingOffset());
	SetPosition(finalPosition);

	return finalPosition;
}

//--------------------------------------------------------------------------------------------------------------------//

const tbMath::Unstable::BoundingVolume PlayerEntity::GetWaterBoundingVolume(void) const
{
	return tbMath::Unstable::BoundingVolume(48.0f, 32.0f);
}

//--------------------------------------------------------------------------------------------------------------------//

const tbMath::Vector2 PlayerEntity::GetWaterBoundingOffset(void) const
{
	if (true == mIsFacingLeft)
	{
		return tbMath::Vector2(-28.0f, 0.0f);
	}
	return tbMath::Vector2(28.0f, 0.0f);
}

//--------------------------------------------------------------------------------------------------------------------//

void PlayerEntity::SimulateUsingWater(void)
{
	const tbMath::Unstable::BoundingVolume& waterBoundingVolume = GetWaterBoundingVolume();
	const tbMath::Vector2& waterBoundingOffset = GetWaterBoundingOffset();

	bool hasCollided(false);
	std::list<EntityInterface*> treeEntities = GetEntityManager()->GetEntitiesByType("TreeEntity");
	for (EntityInterface* entity : treeEntities)
	{
		TreeEntity* tree = dynamic_cast<TreeEntity*>(entity);
		tb_error_if(nullptr == tree, "Expected tree to be an object of type TreeEntity");
		if (false == hasCollided && tree->GetBoundingVolume().CheckCollisionWith(tree->GetPosition() + tree->GetBoundingOffset(), waterBoundingVolume, GetPosition() + waterBoundingOffset))
		{			
			tree->GrowBigger();
			hasCollided = true;
			break;
		}
	}

	std::list<EntityInterface*> fireEntities = GetEntityManager()->GetEntitiesByType("FireEntity");
	for (EntityInterface* entity : fireEntities)
	{
		FireEntity* fire = dynamic_cast<FireEntity*>(entity);
		tb_error_if(nullptr == fire, "Expected tree to be an object of type FireEntity");
		if (false == hasCollided && fire->GetBoundingVolume().CheckCollisionWith(fire->GetPosition() + fire->GetBoundingOffset(), waterBoundingVolume, GetPosition() + waterBoundingOffset))
		{
			fire->WaterOnFire();
			hasCollided = true;
			break;
		}
	}
}

//--------------------------------------------------------------------------------------------------------------------//

void PlayerEntity::OnBurnToDeath(void)
{
	if (false == mIsAlive || true == mHasWonLevel) { return; }

	mIsAlive = false;
	SetCollidable(false);

	if (true == mIsFacingLeft)
	{
		if (false == mSprite.IsPlayingAnimation("death_left")) { mSprite.PlayAnimation("death_left", false, true, kAnimationFrameRate); }
	}
	else
	{
		if (false == mSprite.IsPlayingAnimation("death_right")) { mSprite.PlayAnimation("death_right", false, true, kAnimationFrameRate); }
	}

	tbAudio::theAudioManager.PlayEvent("audio_events", "burn_death");
}

//--------------------------------------------------------------------------------------------------------------------//

void PlayerEntity::OnFallDeath(void)
{
	if (false == mIsAlive || true == mHasWonLevel) { return; }

	mIsAlive = false;
	SetCollidable(false);

	if (true == mIsFacingLeft)
	{
		if (false == mSprite.IsPlayingAnimation("death_left")) { mSprite.PlayAnimation("death_left", false, true, kAnimationFrameRate); }
	}
	else
	{
		if (false == mSprite.IsPlayingAnimation("death_right")) { mSprite.PlayAnimation("death_right", false, true, kAnimationFrameRate); }
	}

	mSprite.SetPosition(tbMath::Vector2(0.0f, 1.0f) + mSprite.GetPosition());	//Hack
	tbAudio::theAudioManager.PlayEvent("audio_events", "fall_death");
}

//--------------------------------------------------------------------------------------------------------------------//

void PlayerEntity::OnSuicidalDeath(void)
{
	if (false == mIsAlive || true == mHasWonLevel) { return; }

	mIsAlive = false;
	SetCollidable(false);

	if (true == mIsFacingLeft)
	{
		if (false == mSprite.IsPlayingAnimation("death_left")) { mSprite.PlayAnimation("death_left", false, true, kAnimationFrameRate); }
	}
	else
	{
		if (false == mSprite.IsPlayingAnimation("death_right")) { mSprite.PlayAnimation("death_right", false, true, kAnimationFrameRate); }
	}

	tbAudio::theAudioManager.PlayEvent("audio_events", "suicide_death");
}

//--------------------------------------------------------------------------------------------------------------------//

void PlayerEntity::ResetToStart(void)
{
	mIsAlive = true;
	mHasWonGame = false;
	mHasWonLevel = false;
	mIsOnTree = false;
	mIsFacingLeft = false;
	mIsMovingLeft = false;
	mIsMovingRight = false;
	mIsClimbingUp = false;
	mIsClimbingDown = false;
	mIsUsingWater = false;
	mIsUsingSeeds = false;	
	mSeedsUseTimer = tbGame::GameTimer(0);
	mWaterUseTimer = tbGame::GameTimer(0);

	mClimbMove = kClimbNone;
	mClimbMoveTimer = tbGame::GameTimer(0);

	const tbGame::AboutTileContainer& startTiles = mTileSystem.FindTilesWithPropertyEquals("player_start", true);
	tb_error_if(startTiles.size() != 1, "Expected TileSystem to have exactly one tile with property player_start");

	SetPosition(startTiles[0].mCenterX, startTiles[0].mTop - 0.1f);
	mPreviousPosition = GetPosition();
	mLastSafeHeight = GetPosition();

	mRemainingSeeds = mTileSystem.GetMapPropertyAsInteger("start_seeds");
	mRemainingWater = mTileSystem.GetMapPropertyAsInteger("start_water");

	if (true == IsOnGround())
	{
		if (true == mIsFacingLeft)
		{
			if (false == mSprite.IsPlayingAnimation("stand_left")) { mSprite.PlayAnimation("stand_left", true, true, kAnimationFrameRate); }
		}
		else
		{
			if (false == mSprite.IsPlayingAnimation("stand_right")) { mSprite.PlayAnimation("stand_right", true, true, kAnimationFrameRate); }
		}
	}
	else //assume starting in a tree
	{
		mSprite.PlayAnimation("climb_up", false, false, kAnimationFrameRate);
		mSprite.StopAnimation();
		mIsOnTree = true;
	}
}

//--------------------------------------------------------------------------------------------------------------------//
